import json
import pathlib

import click
import praw
from prawcore import OAuthException
from selenium.common.exceptions import NoSuchElementException

from crunchy_bot.config.json_config_parser import JsonConfigParser
from crunchy_bot.fetcher.guest_pass_fetcher import GuestPassFetcher
from crunchy_bot.logging.click_logger import ClickLogger
from crunchy_bot.publisher.reddit_publisher import RedditPublisher


class Context(object):
    def __init__(self):
        self.logger = ClickLogger()


@click.group()
@click.pass_context
def cli(context):
    context.obj = Context()


@cli.command()
@click.pass_obj
def init(context):
    """
    Initialize CrunchyBot's credentials file.
    """
    crunchy_username = click.prompt("Crunchyroll Username")
    crunchy_password = click.prompt("Crunchyroll Password", hide_input=True)
    reddit_client_id = click.prompt("Reddit Client ID")
    reddit_client_secret = click.prompt("Reddit Client Secret", hide_input=True)
    reddit_username = click.prompt("Reddit Username")
    reddit_user_agent = click.prompt(
        "Reddit User Agent",
        default=f"CrunchyBot:v4.0.0 (hosted by /u/{reddit_username})",
        show_default=True,
    )
    reddit_password = click.prompt("Reddit Password", hide_input=True)
    log_dir = click.prompt(
        "Log output directory", default="/tmp/crunchybot/logs", show_default=True
    )
    config_path = click.prompt(
        "Where to save config",
        default=pathlib.Path.home().joinpath(".crunchybot").as_posix(),
        show_default=True,
    )

    config = dict(
        crunchy_username=crunchy_username,
        crunchy_password=crunchy_password,
        reddit_client_id=reddit_client_id,
        reddit_client_secret=reddit_client_secret,
        reddit_user_agent=reddit_user_agent,
        reddit_username=reddit_username,
        reddit_password=reddit_password,
        log_dir=log_dir,
    )

    config_path = pathlib.Path(config_path).resolve()
    if config_path.exists():
        if config_path.is_dir():
            context.logger.error(f"{config_path.as_posix()} is a directory")
            exit(1)
        else:
            overwrite = click.confirm(
                f"Overwrite {config_path.as_posix()}", default=False, show_default=True
            )
            if not overwrite:
                exit(1)
    with open(config_path, "w") as f:
        json.dump(config, f, indent=True)


@cli.command()
@click.option(
    "--config",
    type=click.Path(exists=True),
    default=pathlib.Path.home().joinpath(".crunchybot").as_posix(),
)
@click.option("--debug", "-d", type=bool, is_flag=True, default=False)
@click.pass_obj
def publish(context: Context, config: str, debug: bool):
    """
    Searches for Guest Passes and publishes it to reddit.
    """
    logger = context.logger
    data_path = pathlib.Path(config)

    config_parser = JsonConfigParser(logger=logger)
    config = None
    try:
        config = config_parser.parse(data_path)
    except ValueError as e:
        logger.error(f"Failed to parse config: {str(e)}")
        exit(1)

    fetcher = GuestPassFetcher(config, logger=logger)
    guest_passes = []
    try:
        guest_passes.extend(fetcher.fetch(debug=debug))

        if len(guest_passes) == 0:
            logger.info("No valid Guest Passes to publish")
            exit(0)
        logger.info(f"Received Guest Passes: {guest_passes}")
    except NoSuchElementException:
        logger.error(
            "Failed to fetch Guest Passes. Please check your Crunchyroll credentials."
        )
        exit(1)

    client = praw.Reddit(
        client_id=config.reddit_client_id,
        client_secret=config.reddit_client_secret,
        user_agent=config.reddit_user_agent,
        username=config.reddit_username,
        password=config.reddit_password,
    )
    publisher = RedditPublisher(client, logger=logger)

    try:
        logger.info("Publishing Guest Passes to Reddit...")
        if publisher.publish(guest_passes):
            logger.info("Published Guest Passes to Reddit")
        else:
            logger.warn("Failed to find thread to publish to Reddit")
    except OAuthException as e:
        logger.error(f"Failed to login to Reddit: {str(e)}")
        exit(1)



import pathlib


class Config(object):
    def __init__(
        self,
        crunchy_username,
        crunchy_password,
        reddit_client_id,
        reddit_client_secret,
        reddit_user_agent,
        reddit_username,
        reddit_password,
        log_dir,
    ):
        self.crunchy_username = crunchy_username
        self.crunchy_password = crunchy_password
        self.reddit_client_id = reddit_client_id
        self.reddit_client_secret = reddit_client_secret
        self.reddit_user_agent = reddit_user_agent
        self.reddit_username = reddit_username
        self.reddit_password = reddit_password
        self.log_dir = log_dir

    def __eq__(self, other):
        if not isinstance(other, Config):
            return False
        return (
            self.crunchy_username == other.crunchy_username
            and self.crunchy_password == other.crunchy_password
            and self.reddit_client_id == other.reddit_client_id
            and self.reddit_client_secret == other.reddit_client_secret
            and self.reddit_user_agent == other.reddit_user_agent
            and self.reddit_username == other.reddit_username
            and self.reddit_password == other.reddit_password
            and self.log_dir == other.log_dir
        )

    def __hash__(self):
        return hash(
            (
                self.crunchy_username,
                self.crunchy_password,
                self.reddit_client_id,
                self.reddit_client_secret,
                self.reddit_user_agent,
                self.reddit_username,
                self.reddit_password,
                self.log_dir,
            )
        )


class ConfigParser(object):
    def parse(self, path: pathlib.Path) -> Config:
        """
        Given a path, parse and return a Config object with data.

        :param path: Path to configuration file to parse.
        :return: Config object with data from configuration file.
        """
        raise NotImplementedError("parse is not implemented")



import json
import pathlib
from typing import Optional

from crunchy_bot.config.config_parser import ConfigParser, Config
from crunchy_bot.logging.logger import Logger
from crunchy_bot.logging.noop_logger import NoopLogger


class JsonConfigParser(ConfigParser):
    def __init__(self, logger: Optional[Logger] = None):
        self.logger = logger if logger is not None else NoopLogger()

    def parse(self, path: pathlib.Path) -> Config:
        self.logger.info("Fetching Account Data...")
        posix_path = path.resolve().as_posix()
        with open(posix_path, "r") as data_file:
            data_dictionary = json.load(data_file)

        crunchy_username = data_dictionary.get("crunchy_username", None)
        if crunchy_username is None:
            raise ValueError(f"Missing crunchy_username in {posix_path}")
        crunchy_password = data_dictionary.get("crunchy_password", None)
        if crunchy_password is None:
            raise ValueError(f"Missing crunchy_password in {posix_path}")
        reddit_client_id = data_dictionary.get("reddit_client_id", None)
        if reddit_client_id is None:
            raise ValueError(f"Missing reddit_client_id in {posix_path}")
        reddit_client_secret = data_dictionary.get("reddit_client_secret", None)
        if reddit_client_secret is None:
            raise ValueError(f"Missing reddit_client_secret in {posix_path}")
        reddit_user_agent = data_dictionary.get("reddit_user_agent", None)
        if reddit_user_agent is None:
            raise ValueError(f"Missing reddit_user_agent in {posix_path}")
        reddit_username = data_dictionary.get("reddit_username", None)
        if reddit_username is None:
            raise ValueError(f"Missing reddit_username in {posix_path}")
        reddit_password = data_dictionary.get("reddit_password", None)
        if reddit_password is None:
            raise ValueError(f"Missing reddit_password in {posix_path}")
        log_dir = data_dictionary.get("log_dir", None)
        if log_dir is None:
            raise ValueError(f"Missing log_dir in {posix_path}")

        self.logger.info("Fetched Account Data")

        return Config(
            crunchy_username=crunchy_username,
            crunchy_password=crunchy_password,
            reddit_client_id=reddit_client_id,
            reddit_client_secret=reddit_client_secret,
            reddit_user_agent=reddit_user_agent,
            reddit_username=reddit_username,
            reddit_password=reddit_password,
            log_dir=log_dir,
        )



from typing import Callable
from unittest.mock import patch

import click
import pytest

from crunchy_bot.logging.click_logger import ClickLogger
from crunchy_bot.logging.logger import LogColor, Logger, LogTag


@pytest.fixture()
def click_logger():
    return ClickLogger()


def test_success(click_logger: Logger):
    run_test(click_logger.success, "testing", LogTag.SUCCESS, LogColor.SUCCESS)


def test_info(click_logger: Logger):
    run_test(click_logger.info, "testing", LogTag.INFO, LogColor.INFO)


def test_warn(click_logger: Logger):
    run_test(click_logger.warn, "testing", LogTag.WARN, LogColor.WARN)


def test_error(click_logger: Logger):
    run_test(click_logger.error, "testing", LogTag.ERROR, LogColor.ERROR)


def test_critical(click_logger: Logger):
    run_test(click_logger.critical, "testing", LogTag.CRITICAL, LogColor.CRITICAL)


@patch("crunchy_bot.logging.click_logger.click", autospec=True)
def test_log(click_mock: click, click_logger: Logger):
    tag, message, color = LogTag.INFO, "testing", LogColor.CRITICAL
    click_logger.log(tag, message, color)
    click_mock.secho.assert_called_with(f"[ {tag.name[:4]} ] {message}", fg=color.value)


def run_test(
    logging_method: Callable[[str], None], message: str, tag: LogTag, color: LogColor
):
    with patch("crunchy_bot.logging.click_logger.click", autospec=True) as click_mock:
        logging_method(message)
        click_mock.secho.assert_called_with(
            f"[ {tag.name[:4]} ] {message}", fg=color.value
        )



import pathlib
from enum import Enum
from typing import Sequence, Optional

from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from crunchy_bot.config.config_parser import Config
from crunchy_bot.fetcher.fetcher import Fetcher
from crunchy_bot.logging.logger import Logger
from crunchy_bot.logging.noop_logger import NoopLogger


class GuestPassStatus(Enum):
    UNKNOWN = "UNKNOWN"
    VALID = "VALID"
    EXPIRED = "EXPIRED"
    REDEEMED = "REDEEMED"

    @staticmethod
    def to_enum(value: str) -> Enum:
        value = value.upper()
        if value == GuestPassStatus.VALID.value:
            return GuestPassStatus.VALID
        elif value == GuestPassStatus.EXPIRED.value:
            return GuestPassStatus.EXPIRED
        elif value == GuestPassStatus.REDEEMED.value:
            return GuestPassStatus.REDEEMED
        return GuestPassStatus.UNKNOWN


class Row(object):
    """
    Representation of a row of the Guest Pass Table
    """

    def __init__(
        self,
        created_at: str,
        guest_pass: str,
        description: str,
        status: str,
        expiration: str,
        redeemer: str,
        action: str,
    ):
        self.created_at = created_at
        self.guest_pass = guest_pass
        self.description = description
        self.status = GuestPassStatus.to_enum(status)
        self.expiration = expiration
        self.redeemer = redeemer
        self.action = action


class GuestPassFetcher(Fetcher):
    def __init__(self, config: Config, logger: Optional[Logger] = None):
        self.config = config
        self.logger = logger if logger is not None else NoopLogger()

    def fetch(self, debug: bool = False) -> Sequence[str]:
        """
        Fetch CrunchyRoll Guest Passes if any.

        Args:
            debug: Boolean dictating to use ChromeDriver for debugging purposes or not.

        Returns:
            List of Guest Passes.
        """
        self.logger.info("Fetching Guest Passes...")
        guest_passes = self._crunchy_data_fetch(debug=debug)
        self.logger.info("Fetched Guest Passes")

        return guest_passes

    def _crunchy_data_fetch(self, debug=False) -> Sequence[str]:
        """
            Fetch Guest Passes from given CrunchyRoll Account.

            Args:
                debug:      Boolean used to dictate whether to use ChromeDriver or not.
            Returns:
                List of Guest Passes as Strings.
        """
        # Constants.
        GUEST_PASS_TABLE_INDEX = 0
        GUEST_PASS_TABLE_COLUMNS = 7

        # List to be returned. Will hold all valid guest passes.
        valid_guest_passes = []

        log_path = pathlib.Path(self.config.log_dir)
        if not log_path.exists():
            log_path.mkdir(parents=True)
        chrome_options = Options()
        chrome_options.add_argument(
            f"--log-path={log_path.joinpath('chrome.log').as_posix()}"
        )
        if not debug:
            chrome_options.add_argument("--headless")
        driver = webdriver.Chrome(options=chrome_options)

        driver.get("https://www.crunchyroll.com/login?next=%2F")

        # Login to CrunchyRoll
        try:
            # Since CloudFlare stalls the login page, this is to wait the estimated
            # 5 seconds (20 seconds to be sure) for CloudFlare to approve of
            # browser.
            username_field = WebDriverWait(driver, 20).until(
                EC.presence_of_element_located((By.ID, "login_form_name"))
            )
            username_field.send_keys(self.config.crunchy_username)

            password_field = WebDriverWait(driver, 20).until(
                EC.presence_of_element_located((By.ID, "login_form_password"))
            )
            password_field.send_keys(self.config.crunchy_password)
            password_field.send_keys(Keys.ENTER)
        except TimeoutException:
            driver.quit()
            raise TimeoutException(
                "Unable to find username/password field: Crunchyroll took too long to load"
            )

        # Navigate to the last page of the Guest Pass page.~
        driver.get("https://www.crunchyroll.com/acct/?action=guestpass")

        # Grabs HTML data.
        guest_pass_tables = driver.find_elements_by_class_name("acct-guestpass-tl")

        # Ensure user was able to login.
        if not guest_pass_tables:
            raise NoSuchElementException

        row_list = guest_pass_tables[GUEST_PASS_TABLE_INDEX].find_elements_by_tag_name(
            "tr"
        )

        # Parse HTML table data.
        for row in row_list:
            cell_list = row.find_elements_by_tag_name("td")
            if len(cell_list) == GUEST_PASS_TABLE_COLUMNS:
                cell_list = list(map(lambda cell: cell.text, cell_list))
                row_entry = Row(*cell_list)
                if row_entry.status is GuestPassStatus.VALID:
                    valid_guest_passes.append(row_entry.guest_pass)
                    break

        # Close the driver.
        driver.close()
        return valid_guest_passes



from typing import Sequence


class Publisher(object):
    def publish(self, guest_passes: Sequence[str]) -> bool:
        """
        Given a sequence of guest passes, publish them to some location

        :param guest_passes: Sequence of Guest Pass strings.
        :return: True if and only if it was successful.
        """
        raise NotImplementedError("publish is not implemented")



import os
import pathlib

from setuptools import setup, find_packages

setup_path = pathlib.Path(os.path.dirname(os.path.relpath(__file__)))
with open(setup_path.joinpath("README.md").as_posix(), encoding="utf-8", mode="r") as f:
    long_description = f.read()

setup(
    name="crunchy_bot",
    use_scm_version=True,
    setup_requires=["setuptools_scm", "pytest-runner"],
    description="Crunchyroll Guest Pass Publisher for Reddit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="David Lam",
    author_email="david.lam@lamdav.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: MacOS X",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
    ],
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    install_requires=[
        "Click==7.0",
        "praw==6.1.0",
        "prawcore==1.0.0",
        "selenium==3.141.0",
    ],
    tests_require=["pytest"],
    python_requires=">=3.5",
    entry_points={"console_scripts": ["crunchy=crunchy_bot.client.cli:cli"]},
    include_package_data=True,
    project_urls={
        "Bug Reports": "https://github.com/lamdaV/CrunchyBot/issues",
        "Source": "https://github.com/lamdaV/CrunchyBot",
    },
)



from unittest import mock

import praw
import pytest

from crunchy_bot.logging.logger import Logger
from crunchy_bot.publisher.reddit_publisher import RedditPublisher


@pytest.fixture()
def client():
    return mock.MagicMock(name="reddit_mock")


@pytest.fixture()
def logger():
    return mock.create_autospec(Logger)


def test_publish(client: praw.Reddit, logger: Logger):
    username = "reddit_username"
    client.user.me.return_value = username
    subreddit_mock = mock.MagicMock(name="subreddit_mock")
    client.subreddit.return_value = subreddit_mock
    submission_mock = mock.MagicMock()
    submission_mock.title = "Unrelated post"
    guest_pass_submission_mock = mock.MagicMock()
    guest_pass_submission_mock.title = (
        "Weekly Guest Pass MegaThread for the week of May 27, 2019, "
        "Please use this thread to ask or give away Guest Passes."
    )
    subreddit_mock.hot.return_value = [submission_mock, guest_pass_submission_mock]

    publisher = RedditPublisher(client, logger=logger)
    guest_passes = ["123", "abc", "foo"]
    assert publisher.publish(guest_passes)

    logger.info.assert_any_call("Building Comment Text...")
    logger.info.assert_any_call("Built Comment Text")
    logger.info.assert_any_call(f"Logged in as {username}...")
    client.user.me.assert_any_call()
    client.subreddit.assert_any_call("Crunchyroll")
    guest_pass_submission_mock.reply.assert_any_call(
        "Here are some valid passes:  \n\n"
        " * 123\n"
        " * abc\n"
        " * foo\n"
        "  \n*Disclaimer: This is a bot. Here is a [link](https://github.com/lamdaV/CrunchyBot/tree/master) for more detail.*"
    )
